#include <mad.h>

#include <nds.h>

#include <stdio.h>
#include <string.h>

#include <ipc2.h>

#define SOUND_NEED_BUFFER

#include "sound.h"
#include "file.h"

#define INPUT_BUFFER_SIZE 2048
#define FILE_BUFFER_SIZE 4096

static struct mad_stream Stream;
static struct mad_frame Frame;
static struct mad_synth Synth;
static mad_timer_t Timer;
static unsigned char *InputBuffer, *GuardPtr;
static unsigned FrameCount;

static char buffer[FILE_BUFFER_SIZE];
static u32 buflen = 0;

void madplay_stop(void) {
	mad_synth_finish(&Synth);
	mad_frame_finish(&Frame);
	mad_stream_finish(&Stream);
}

static s16 madplay_fixedtos16(mad_fixed_t sample) {
	sample += (1L << (MAD_F_FRACBITS - 16));

	/* Clipping */
	if(sample > MAD_F_ONE - 1)
		sample = MAD_F_ONE - 1;
	if(sample < -MAD_F_ONE)
		sample = MAD_F_ONE;

	/* Conversion. */
	return sample >> (MAD_F_FRACBITS + 1 - 16);
}

static int madplay_read(void *ptr, u32 size) {
	u32 read;

	if(buflen < FILE_BUFFER_SIZE) {
		buflen += fread(buffer + buflen, 1, FILE_BUFFER_SIZE - buflen, sound_playfile);
	}

	if(buflen < size) {
		memcpy(ptr, buffer, buflen);
		read = buflen;
		buflen = 0;
	} else {
		memcpy(ptr, buffer, size);
		memmove(buffer, buffer + size, FILE_BUFFER_SIZE - size);
		read = size;
		buflen -= size;
	}

	if(buflen < FILE_BUFFER_SIZE) {
		buflen += fread(buffer + buflen, 1, FILE_BUFFER_SIZE - buflen, sound_playfile);
	}

	return read;
}

static int madplay_eof(void) {
	if(buflen == 0)
		return feof(sound_playfile);
	else
		return 0;
}

static int madplay_decode(void) {
	int i;
	s16 *madpcmL = (s16 *) pcmL, *madpcmR = (s16 *) pcmR;

	if(state == FINISHING)
		return 0;

	if(Stream.buffer == NULL || Stream.error == MAD_ERROR_BUFLEN) {
		size_t ReadSize, Remaining;
		unsigned char *ReadStart;

		if(Stream.next_frame != NULL) {
			Remaining = Stream.bufend - Stream.next_frame;
			memmove(InputBuffer, Stream.next_frame, Remaining);
			ReadStart = InputBuffer + Remaining;
			ReadSize = INPUT_BUFFER_SIZE - Remaining;
		} else {
			ReadSize = INPUT_BUFFER_SIZE;
			ReadStart = InputBuffer;
			Remaining = 0;
		}

		madplay_read(ReadStart, ReadSize);
		if(madplay_eof()) {
			GuardPtr = ReadStart + ReadSize;
			memset(GuardPtr, 0, MAD_BUFFER_GUARD);
			ReadSize += MAD_BUFFER_GUARD;

			state = FINISHING;
		}

		mad_stream_buffer(&Stream, InputBuffer, ReadSize + Remaining);
		Stream.error = 0;
	}

	if(mad_frame_decode(&Frame, &Stream)) {
		if(MAD_RECOVERABLE(Stream.error)) {
			if(Stream.error != MAD_ERROR_LOSTSYNC || Stream.this_frame != GuardPtr) {
				if(IPC2->messageflag < IPC2_MAX_MESSAGES)
					sprintf(IPC2->message[IPC2->messageflag++], "recoverable frame level error (%s)\n", mad_stream_errorstr(&Stream));
			}
			return 1;
		} else {
			if(Stream.error == MAD_ERROR_BUFLEN)
				return 1;
			else {
				if(IPC2->messageflag < IPC2_MAX_MESSAGES)
					sprintf(IPC2->message[IPC2->messageflag++], "unrecoverable frame level error (%s).\n", mad_stream_errorstr(&Stream));

				return -1;
			}
		}
	}

	FrameCount++;
	
	mad_timer_add(&Timer, Frame.header.duration);

	mad_synth_frame(&Synth, &Frame);

	for(i=0; i < Synth.pcm.length; i++) {

		madpcmL[i] = madplay_fixedtos16(Synth.pcm.samples[0][i]);

		if(MAD_NCHANNELS(&Frame.header)==2)
			madpcmR[i] = madplay_fixedtos16(Synth.pcm.samples[1][i]);
	}

	IPC2->sound_writerequest = 0;

	return 0;
}

static void InterruptHandler_IPC_SYNC_mad(void) {
	u8 sync;
	int status;

	sync = IPC_GetSync();

	if(sync == IPC2_REQUEST_WRITE_SOUND) {
		switch(format) {
			case MAD:
				while((status = madplay_decode()) != 0);
				if(status < 0)
					sound_stop();

				DC_FlushRange(pcmL, MAD_NCHANNELS(&Frame.header) * 2);
				DC_FlushRange(pcmR, MAD_NCHANNELS(&Frame.header) * 2);

				break;

			default:
				break;
		}
	}
}

int madplay(FILE *fp_arg) {
	int status;
	size_t size;

	InputBuffer = sound_buffer;
	sound_playfile = fp_arg;
	format = MAD;

	mad_stream_init(&Stream);
	mad_frame_init(&Frame);
	mad_synth_init(&Synth);
	mad_timer_reset(&Timer);

	buflen = 0;

	fseek(sound_playfile, 0, SEEK_END);
	size = ftell(sound_playfile);
	fseek(sound_playfile, size/2, SEEK_SET);

	while((status = madplay_decode()) != 0);
	if(status < 0) {
		if(IPC2->messageflag < IPC2_MAX_MESSAGES)
			strcpy(IPC2->message[IPC2->messageflag++], "mad: unrecoverable error");

		madplay_stop();

		return -1;
	}

	IPC2->sound_channels = MAD_NCHANNELS(&Frame.header);
	IPC2->sound_frequency = Frame.header.samplerate;
	IPC2->sound_bytes_per_sample = 2;
	IPC2->sound_samples = 32 * MAD_NSBSAMPLES(&Frame.header);
	IPC2->sound_writerequest = 0;

	mad_synth_finish(&Synth);
	mad_frame_finish(&Frame);
	mad_stream_finish(&Stream);

	mad_stream_init(&Stream);
	mad_frame_init(&Frame);
	mad_synth_init(&Synth);
	mad_timer_reset(&Timer);

	fseek(sound_playfile, 0, SEEK_SET);
	buflen = 0;
	FrameCount = 0;

	IPC2->sound_lbuf = (void *) pcmL;
	IPC2->sound_rbuf = (void *) pcmR;

	irqSet(IRQ_IPC_SYNC, InterruptHandler_IPC_SYNC_mad);
	REG_IPC_SYNC = IPC_SYNC_IRQ_ENABLE;

	return 0;
}

void madplay_playpause(void) {
	if(state == PLAYING)
		IPC_SendSync(IPC2_REQUEST_STOP_PLAYING);
	else if(state == STOPPED || state == PAUSED);
		IPC2->sound_control = IPC2_SOUND_START;
}
